import os
import shutil
from typing import List, Union


def _has_same_content(path1: str, path2: str):
    if os.path.exists(path1):
        with open(path1, 'rb') as f:
            c1 = f.read()
    else:
        c1 = ''
    if os.path.exists(path2):
        with open(path2, 'rb') as f:
            c2 = f.read()
    else:
        c2 = ''
    return c1 == c2

class SyncTool:
    def __init__(self, projects: List[str]):
        self._projects = projects
        self._discrepancies = []
    def scan_dir(self, dirpath: str):
        fnames: List[str] = []
        dirnames: List[str] = []
        for p in self._projects:
            project_name = os.path.basename(p)
            fulldirpath = f'{p}/{dirpath}'.replace('$$projectname$$', project_name)
            if os.path.isdir(fulldirpath):
                for fname in os.listdir(fulldirpath):
                    if not _excluded(fname):
                        fullpath = f'{fulldirpath}/{fname}'
                        if os.path.isfile(fullpath):
                            fnames.append(fname)
                        elif os.path.isdir(fullpath):
                            dirnames.append(fname.replace(project_name, '$$projectname$$'))
        fnames = sorted(list(set(fnames)))
        dirnames = sorted(list(set(dirnames)))
        for fname in fnames:
            relpath = f'{dirpath}/{fname}'
            fullpaths: List[str] = []
            for p in self._projects:
                project_name = os.path.basename(p)
                fullpath = f'{p}/{relpath}'.replace('$$projectname$$', project_name)
                # if os.path.isfile(fullpath):
                fullpaths.append(fullpath)
            # if len(fullpaths) > 1:
            fullpaths = [a for a in fullpaths if not _file_is_jinjaroot_generated(a)]
            groups = _split_into_groups(fullpaths)
            newest_path = _get_newest_path(fullpaths)
            if len(groups) > 1:
                self._discrepancies.append({'relpath': relpath, 'groups': groups, 'newest_path': newest_path})
        for dirname in dirnames:
            relpath = f'{dirpath}/{dirname}'
            self.scan_dir(relpath)
    def report(self, *, ask: bool):
        for discr in self._discrepancies:
            print('***********************************************************')
            print(f'*** {discr["relpath"]}')
            groups = discr['groups']
            newest_path = discr['newest_path']
            something_missing = False
            for g in groups:
                for f in g:
                    x: List[str] = []
                    if not os.path.exists(f):
                        x.append('missing')
                        something_missing = True
                    if f == newest_path:
                        x.append('newest')
                    if len(x) == 0:
                        print(f)
                    else:
                        print(f'{f} ({", ".join(x)})')
                print('')
            if ask:
                print('')
                if something_missing:
                    reply = _prompt('What would you like to do? s=skip, n=use newest, d=delete all: ', ['s', 'n', 'd'])
                else:
                    reply = _prompt('What would you like to do? s=skip, n=use newest: ', ['s', 'n'])
                if reply == 's':
                    pass
                elif reply == 'n':
                    for g in groups:
                        for f in g:
                            if f != newest_path:
                                print(f'Copying {newest_path} {f}')
                                shutil.copyfile(newest_path, f)
                elif reply == 'd':
                    for g in groups:
                        for f in g:
                            if os.path.exists(f):
                                if _prompt(f'Delete {f}? ', ['y', 'n']) == 'y':
                                    print(f'Deleting {f}')
                                    os.unlink(f)
            print('')

def _prompt(question: str, valid_answers: List[str]):
    answer = None
    while answer not in valid_answers:
        answer = str(input(question))
    return answer

def _get_newest_path(paths: List[str]):
    ret = ''
    newest_mtime: Union[float, None] = None
    for path in paths:
        if os.path.exists(path):
            mtime = os.stat(path).st_mtime
            if newest_mtime is None or (mtime > newest_mtime):
                newest_mtime = mtime
                ret = path
    return ret

def _file_is_jinjaroot_generated(path: str):
    if not os.path.exists(path):
        return False
    with open(path, 'rb') as f:
        c = f.read()
    return (b'This file was automatically generated by jinjaroot.' in c) or (b'jinjaroot synctool exclude' in c)

def _split_into_groups(paths: List[str]):
    groups: List[List[str]] = []
    for path in paths:
        found = False
        for group in groups:
            if not found:
                if _has_same_content(group[0], path):
                    group.append(path)
                    found = True
        if not found:
            groups.append([path])
    return groups

def _excluded(fname: str):
    if fname.endswith('.egg-info'): return True
    if fname.endswith('.gen.ts'): return True
    return fname in ['__pycache__', 'node_modules', '.git', '.vercel', 'tmp', 'yarn.lock', 'package.json', 'jinjaroot.yaml']

def synctool(projects, *, ask: bool):
    s = SyncTool(projects)
    s.scan_dir('.')
    s.report(ask=ask)



import os
from typing import Union
import yaml
import json
from jinja2 import Template
from copy import deepcopy

header_msg1 = 'This file was automatically generated by jinjaroot. Do not edit directly. See the .jinjaroot dir.'
header_msg2 = 'This file was automatically generated by jinjaroot. Do not edit directly.'

def process_jinjaroot_dir(*, src_dirpath: str, dst_dirpath: str, template_data: dict, verify_only: bool, dry: bool):
    template_fnames = os.listdir(src_dirpath)
    for fname in template_fnames:
        template_path = f'{src_dirpath}/{fname}'
        fname2 = Template(fname).render(**template_data)
        dest_path = f'{dst_dirpath}/{fname2}'
        if os.path.isdir(template_path):
            if not os.path.exists(dest_path):
                os.mkdir(dest_path)
            process_jinjaroot_dir(src_dirpath=template_path, dst_dirpath=dest_path, template_data=template_data, verify_only=verify_only, dry=dry)
        elif os.path.isfile(template_path):
            with open(template_path, 'r') as f:
                template_code = f.read()
            t = Template(template_code, keep_trailing_newline=True)
            gen_code = t.render(**template_data)
            gen_code = add_gen_header(fname, gen_code, header_msg1)
            _check_and_write_file(dst_path=dest_path, gen_code=gen_code, verify_only=verify_only, dry=dry, template_path=template_path)

def _check_and_write_file(*, dst_path: str, gen_code: str, verify_only: bool, dry: bool, template_path: Union[str, None]):
    if template_path is not None:
        template_stat = os.stat(template_path)
    else:
        template_stat = None
    if os.path.exists(dst_path):
        with open(dst_path, 'r') as f:
            dst_code = f.read()
        dst_stat = os.stat(dst_path)
    else:
        dst_code = None
        dst_stat = None
    if gen_code != dst_code:
        if verify_only:
            raise Exception(f'Code generation is not up-to-date for file: {dst_path}')
        if dry:
            print(f'Will update: {dst_path}')
        else:
            print(f'Writing: {dst_path}')
            with open(dst_path, 'w') as f:
                f.write(gen_code)
            if (dst_stat is None) or (template_stat and (template_stat.st_mode != dst_stat.st_mode)):
                os.chmod(dst_path, template_stat.st_mode)

def process_dir(*, dirpath: str, template_data: dict, verify_only: bool, dry: bool):
    jinja_root_yaml_path = f'{dirpath}/jinjaroot.yaml'
    jinja_root_json_path = f'{dirpath}/jinjaroot.json'
    jinjaroot_dirpath = f'{dirpath}/.jinjaroot'
    template_data2 = deepcopy(template_data)
    if os.path.isfile(jinja_root_yaml_path):
        with open(jinja_root_yaml_path, 'r') as f:
            x = yaml.safe_load(f)
            for k, v in x.items():
                template_data2[k] = v
    if os.path.isfile(jinja_root_json_path):
        with open(jinja_root_json_path, 'r') as f:
            x = json.load(f)
            for k, v in x.items():
                template_data2[k] = v
    if os.path.isdir(jinjaroot_dirpath):
        process_jinjaroot_dir(src_dirpath=jinjaroot_dirpath, dst_dirpath=dirpath, template_data=template_data2, verify_only=verify_only, dry=dry)
    
    fnames = os.listdir(dirpath)
    for fname in fnames:
        file_path = f'{dirpath}/{fname}'
        if os.path.isfile(file_path):
            template_path = file_path + '.j2'
            if os.path.exists(template_path):
                with open(template_path, 'r') as f:
                    template_code = f.read()
                t = Template(template_code, keep_trailing_newline=True)
                gen_code = t.render(**template_data2)
                gen_code = add_gen_header(fname, gen_code, header_msg2)
                _check_and_write_file(dst_path=file_path, gen_code=gen_code, verify_only=verify_only, dry=dry, template_path=template_path)
            if ends_with(file_path, ['.json']):
                substitute_template_path = file_path + '.substitute.j2'
                if os.path.exists(substitute_template_path):
                    with open(file_path, 'r') as f:
                        content1 = json.load(f)
                    with open(substitute_template_path, 'r') as f:
                        txt = f.read()
                        txt2 = Template(txt, keep_trailing_newline=True).render(**template_data2)
                        content2 = json.loads(txt2)
                    content1_new = _substitute_json(content1, content2)
                    gen_code = json.dumps(content1_new, indent=2)
                    _check_and_write_file(dst_path=file_path, gen_code=gen_code, verify_only=verify_only, dry=dry, template_path=None)
        elif os.path.isdir(file_path):
            process_dir(dirpath=file_path, template_data=template_data2, verify_only=verify_only, dry=dry)

def _substitute_json(content1: dict, content2: dict):
    ret = deepcopy(content1)
    for k, v in content2.items():
        ret[k] = v
    return ret

def ends_with(fname, suffixes):
    for suf in suffixes:
        if fname.endswith(suf): return True
    return False

def insert_header(header: str, code: str):
    lines = code.splitlines()
    line1 = lines[0] if len(lines) > 0 else ''
    if line1.startswith('#!'):
        return code.replace(line1, line1 + '\n' + header)
    else:
        return header + code

def add_gen_header(fname: str, code: str, message: str):
    if ends_with(fname, ['.js', '.jsx', '.ts', '.tsx']):
        return insert_header(f'// {message}\n', code)
    elif ends_with(fname, ['.py', '.sh', '.yaml', '.yml', '.cfg']):
        return insert_header(f'# {message}\n', code)
    elif ends_with(fname, ['.html']):
        return insert_header(f'<!-- {message} -->\n', code)
    elif ends_with(fname, ['.md']):
        return insert_header(f'[//]: # "{message}"\n', code)
    else:
        return code


def generate(*, verify_only: bool=False, dry: bool=False):
    template_data = {}
    process_dir(dirpath='.', template_data=template_data, verify_only=verify_only, dry=dry)

def verify():
    generate(verify_only=True, dry=False)
